/*
 *  SineGenerator.cpp - Simple sine wave generator
 *
 *  Written by Christian Bauer, Public Domain
 */

#include <AppKit.h>
#include <InterfaceKit.h>
#include <MediaKit.h>
#include <math.h>

#include "TSliderView.h"


// Sample frequency in Hz
const float SAMPLE_FREQ = 44100.0f;


// Constants
const char APP_SIGNATURE[] = "application/x-vnd.cebix-SineGenerator";
const uint32 MSG_DIVISOR = 'divi';
const uint32 MSG_RESET_PHASE = 'rsph';
const rgb_color fill_color = {216, 216, 216, 0};


// Main window object
class MainWindow : public BWindow {
public:
	MainWindow();
	virtual bool QuitRequested(void);
	virtual void MessageReceived(BMessage *msg);

private:
	static void ampl_callback(float val, void *arg);
	static void freq_callback(float val, void *arg);
	static void playbuffer_func(void *arg, void *buf, size_t size, const media_raw_audio_format &format);

	TSliderView *ampl_slider[4];
	TSliderView *freq_slider[4];

	BSoundPlayer *the_player;

	float phi0, phi1, phi2, phi3;	// Phase angles
	float divisor;					// Highest output frequency = SAMPLE_FREQ / divisor
	bool reset_phase;				// Flag: reset phase angles
};

static MainWindow *the_window = NULL;


// Application object
class SineGenerator : public BApplication {
public:
	SineGenerator() : BApplication(APP_SIGNATURE) {}
	virtual void ReadyToRun(void) {new MainWindow;}
	virtual void AboutRequested(void);
};


/*
 *  Create application object and start it
 */

int main(int argc, char **argv)
{
	SineGenerator *the_app = new SineGenerator();
	the_app->Run();
	delete the_app;
	return 0;
}


/*
 *  About requested
 */

void SineGenerator::AboutRequested(void)
{
	BAlert *the_alert = new BAlert("", "Sine generator by Christian Bauer\n<cbauer@iphcip1.physik.uni-mainz.de>\nPublic domain.", "Neat");
	the_alert->Go();
}


/*
 *  Window creator
 */

MainWindow::MainWindow() : BWindow(BRect(0, 0, 599, 159), "SineGenerator", B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE)
{
	Lock();
	BRect b = Bounds();
	the_window = this;

	// Move window to right position
	MoveTo(80, 80);

	// Light gray background
	BView *top = new BView(BRect(0, 0, b.right, b.bottom), "", B_FOLLOW_NONE, B_WILL_DRAW);
	AddChild(top);
	top->SetViewColor(fill_color);

	// Labels
	BStringView *label = new BStringView(BRect(10, 10, 75, 20), "ampl_label", "Amplitude");
	top->AddChild(label);
	label = new BStringView(BRect(80, 10, 579, 20), "freq_label", "Frequency");
	top->AddChild(label);

	// Amplitude/frequency sliders
	divisor = 4.0;
	for (int i=0; i<4; i++) {

		float ampl = i ? 0.0 : 0.5;
		float freq = 1000.0 / (SAMPLE_FREQ / divisor);

		// Amplitude display
		BStringView *ampl_display = new BStringView(BRect(10, 38+i*30, 75, 48+i*30), "ampl_display", "");
		top->AddChild(ampl_display);
	
		// Amplitude slider
		ampl_slider[i] = new TSliderView(BRect(10, 20+i*30, 75, 37+i*30), "amplitude", ampl, ampl_callback, ampl_display);
		top->AddChild(ampl_slider[i]);
		ampl_callback(ampl, ampl_display);
	
		// Frequency display
		BStringView *freq_display = new BStringView(BRect(80, 38+i*30, 200, 48+i*30), "freq_display", "");
		top->AddChild(freq_display);
	
		// Frequency slider
		freq_slider[i] = new TSliderView(BRect(80, 20+i*30, 579, 37+i*30), "frequency", freq, freq_callback, freq_display);
		top->AddChild(freq_slider[i]);
		freq_callback(freq, freq_display);
	}
	reset_phase = true;

	// Divisor control
	BCheckBox *box = new BCheckBox(BRect(80, 140, 200, 155), "divisor", "/10", new BMessage(MSG_DIVISOR));
	top->AddChild(box);

	// Reset phase button
	BButton *button = new BButton(BRect(480, 130, 579, 155), "reset_phase", "Reset Phase", new BMessage(MSG_RESET_PHASE));
	top->AddChild(button);

	// Create and set up sound player
	media_raw_audio_format format;
	format.frame_rate = SAMPLE_FREQ;
	format.channel_count = 2;
	format.format = media_raw_audio_format::B_AUDIO_FLOAT;
	format.byte_order = B_HOST_IS_LENDIAN ? B_MEDIA_LITTLE_ENDIAN : B_MEDIA_BIG_ENDIAN;
	format.buffer_size = B_PAGE_SIZE;
	the_player = new BSoundPlayer(&format, "Sine Generator", playbuffer_func, NULL, this);
	if (the_player->Start() != B_NO_ERROR) {
		PostMessage(B_QUIT_REQUESTED);
		return;
	}
	the_player->SetHasData(true);

	// Show window
	Show();
	Unlock();
}


/*
 *  Quit requested, close all and quit program
 */

bool MainWindow::QuitRequested(void)
{
	// Delete sound player
	the_player->Stop();
	delete the_player;

	// Quit program
	be_app->PostMessage(B_QUIT_REQUESTED);
	return true;
}


/*
 *  Message received
 */

void MainWindow::MessageReceived(BMessage *msg)
{
	switch (msg->what) {
		case MSG_DIVISOR:
			if (divisor != 4.0)
				divisor = 4.0;
			else
				divisor = 40.0;
			for (int i=0; i<4; i++)
				freq_slider[i]->SetValue(freq_slider[i]->Value());
			break;

		case MSG_RESET_PHASE:
			reset_phase = true;
			break;

		default:
			BWindow::MessageReceived(msg);
	}
}


/*
 *  Frequency slider moved, display frequency
 */

void MainWindow::freq_callback(float val, void *arg)
{
	char str[80];
	sprintf(str, "%d Hz", (int)(val * SAMPLE_FREQ / the_window->divisor));
	((BStringView *)arg)->SetText(str);
}


/*
 *  Frequency slider moved, display frequency
 */

void MainWindow::ampl_callback(float val, void *arg)
{
	char str[80];
	sprintf(str, "%3.2f", val);
	((BStringView *)arg)->SetText(str);
}


/*
 *  Stream function 
 */

void MainWindow::playbuffer_func(void *arg, void *buf, size_t size, const media_raw_audio_format &format)
{
	MainWindow *obj = (MainWindow *)arg;
	float *p = (float *)buf;

	// Read parameters
	float ampl0 = obj->ampl_slider[0]->Value();
	float freq0 = obj->freq_slider[0]->Value() * SAMPLE_FREQ / obj->divisor;
	float ampl1 = obj->ampl_slider[1]->Value();
	float freq1 = obj->freq_slider[1]->Value() * SAMPLE_FREQ / obj->divisor;
	float ampl2 = obj->ampl_slider[2]->Value();
	float freq2 = obj->freq_slider[2]->Value() * SAMPLE_FREQ / obj->divisor;
	float ampl3 = obj->ampl_slider[3]->Value();
	float freq3 = obj->freq_slider[3]->Value() * SAMPLE_FREQ / obj->divisor;

	if (obj->reset_phase) {
		obj->phi0 = obj->phi1 = obj->phi2 = obj->phi3 = 0.0f;
		obj->reset_phase = false;
	}

	// Read phases
	float phi0 = obj->phi0;
	float phi1 = obj->phi1;
	float phi2 = obj->phi2;
	float phi3 = obj->phi3;

	size >>= 3;		// Float stereo output
	while (size--) {

		// Mix sinus waves and write to buffer
		float val = ampl0 * sin(phi0) + ampl1 * sin(phi1) + ampl2 * sin(phi2) + ampl3 * sin(phi3);
		*p++ = val;
		*p++ = val;

		// Update phase angles
		phi0 += 2.0f * M_PI * freq0 / SAMPLE_FREQ;
		while (phi0 > 2.0f * M_PI)
			phi0 -= 2.0f * M_PI;
		phi1 += 2.0f * M_PI * freq1 / SAMPLE_FREQ;
		while (phi1 > 2.0f * M_PI)
			phi1 -= 2.0f * M_PI;
		phi2 += 2.0f * M_PI * freq2 / SAMPLE_FREQ;
		while (phi2 > 2.0f * M_PI)
			phi2 -= 2.0f * M_PI;
		phi3 += 2.0f * M_PI * freq3 / SAMPLE_FREQ;
		while (phi3 > 2.0f * M_PI)
			phi3 -= 2.0f * M_PI;
	}

	// Write back phases
	obj->phi0 = phi0;
	obj->phi1 = phi1;
	obj->phi2 = phi2;
	obj->phi3 = phi3;
}
